@@ -37,6 +37,7 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 #include <pwd.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#define SYSLOG_NAMES
 #include <sys/syslog.h>
 #include <sys/socket.h>
 #include <sys/event.h>
@@ -45,13 +46,19 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 #include <netinet/ip.h>
 #include <netinet/ip6.h>
 #include <netinet/tcp.h>
+#include <capsicum_helpers.h>
+#include <libcasper.h>
+#include <casper/cap_syslog.h>
 #include "fw.h"
 
+cap_channel_t *capsyslog;
+
 #define LOGX(level, msg, params ...) do { \
   if (log_to_syslog) \
-    syslog(level, "[%s:%03i] " msg "\n", __FILE__, __LINE__, ## params); \
+    cap_syslog(capsyslog, level, "[%s:%03i] " msg "\n", __FILE__, __LINE__, \
+                                                                ## params); \
   else \
-    fprintf(stderr, "[%s:%03i] " msg "\n", __FILE__, __LINE__, ## params); \
+    fprintf(stderr, "[%s:%03i] " msg "\n", __FILE__, __LINE__, ## params);  \
 } while (0)
 
 #define ERRDIEX(msg, params ...) do { \
@@ -259,32 +266,55 @@ void parse_v6(void *buf, unsigned int len, struct rule *rule)
 
 void print_help(FILE *f)
 {
-  fputs(
-"ftp6proxy V0.2 - IPv6 FTP firewall helper\n\n"
-"usage: ftp6proxy [-b <base>] [-c <count>] [-p <port>] [-t <timeout>]\n"
-"                 [-q <queue>] [-d [-d]] [-i <pidfile>] [-u <user>]\n"
-"       ftp6proxy -h\n\n"
+  fprintf(f,
+"ftp6proxy V0.3 - IPv6/IPv4 FTP firewall helper\n"
+"\n"
+"compiled-in firewall module: %s\n"
+"\n"
+"usage: ftp6proxy [-b <base>] [-c <count>] [-d [-d]] [-f <facility>]\n"
+"                 [-i <pidfile>] [-j <target>] [-p <port>] [-q <queue>]\n"
+"                 [-r <dir>] [-t <timeout>] [-u <user>]\n"
+"       ftp6proxy -h\n"
+"\n"
 "  -b <base>      punch holes into firewall starting at this rule number;\n"
 "                 default: 64000\n"
 "  -c <count>     use up to <count> firewall rules; default: 100\n"
-"  -p <port>      divert port number to listen on; defaults to 10004\n"
+"  -d             run in foreground, don't write pidfile\n"
+"  -d -d          run in foreground, don't write pidfile, don't chroot,\n"
+"                 debug log to stderr, don't drop privileges\n"
+"  -f <facility>  send syslog to <facility>; default: daemon\n"
+"  -i <pidfile>   write pid to <pidfile> instead of /var/run/ftp6proxy.pid\n"
+"  -j <target>    call this ipfw rule number instead of allowing the data\n"
+"                 connection\n"
+"  -p <port>      divert(4) port number to listen on; defaults to 10004\n"
+"  -q <queue>     place data connections into this pf QoS class\n"
+"  -r <dir>       chroot into <dir> instead of /var/empty\n"
 "  -t <timeout>   wait up to <timeout> seconds for data connections;\n"
 "                 defaults to 2 seconds\n"
-"  -q <queue>     place data connections into this QoS class\n"
-"  -d             run in foreground, don't write pidfile\n"
-"  -d -d          run in foreground, don't write pidfile, "
-                                                      "debug log to stderr,\n"
-"                 don't drop privileges\n"
-"  -i <pidfile>   write pid to <pidfile> instead of /var/run/ftp6proxy.pid\n"
 "  -u <user>      run as user <user> instead of proxy\n"
-"  -h             show this help ;-)\n\n", f);
+"  -h             show this help ;-)\n"
+"\n", fw_ident());
+}
+
+int syslog_name_to_facility(char *name)
+{
+  unsigned int i;
+
+  for (i = 0; i < sizeof(facilitynames)/sizeof(facilitynames[0]); i++) {
+    if (!strcmp(name, facilitynames[i].c_name))
+      break;
+  }
+
+  return facilitynames[i].c_val;
 }
 
 int main(int argc, char **argv)
 {
   char ch;
   char *pidfile = "/var/run/ftp6proxy.pid";
+  char *chroot_dir = "/var/empty";
   char *user = "proxy";
+  int facility = LOG_DAEMON;
   gid_t gid;
   uid_t uid;
   struct passwd *pw;
@@ -312,37 +342,55 @@ int main(int argc, char **argv)
   struct kevent kev;
   struct timespec ktimeout;
   int numkevents;
+  cap_rights_t rights;
+  cap_channel_t *capcas;
 
-  rule.queue = NULL;
+  rule.pf_queue = NULL;
+  rule.ipfw_call = 0;
 
-  while ((ch = getopt(argc, argv, "b:c:p:t:q:dhi:u:")) != -1) {
+  while ((ch = getopt(argc, argv, "b:c:df:i:j:p:q:r:t:u:h")) != -1) {
     switch (ch) {
       case 'b':
         rule_base = atoi(optarg);
-        if (rule_base < 0) ERRDIEX("rule base must be 0 or greater");
+        if (rule_base <= 0) ERRDIEX("rule base must be 1 or greater");
         break;
       case 'c':
         rule_count = atoi(optarg);
         if (rule_count <= 0) ERRDIEX("rule count must be 1 or greater");
         break;
+      case 'd':
+        debug++;
+        break;
+      case 'f':
+        facility = syslog_name_to_facility(optarg);
+        if (facility < 0) ERRDIEX("unknown syslog facility: %s", optarg);
+        break;
+      case 'i':
+        pidfile = optarg;
+        break;
+      case 'j':
+        rule.ipfw_call = strtoul(optarg, NULL, 10);
+        if ((rule.ipfw_call <= 0) || (rule.ipfw_call >= 65535))
+          ERRDIEX("ipfw target rule number must be in the range from 1 to "
+                                                        "65534, inclusively");
+        break;
       case 'p':
         /* divert port */
         port = atoi(optarg);
         if (port < 0) ERRDIEX("port must be 0 or greater");
         break;
+      case 'q':
+        rule.pf_queue = optarg;
+        if (strlen(rule.pf_queue) == 0)
+          ERRDIEX("pf queue name must not be emtpy");
+        break;
+      case 'r':
+        chroot_dir = optarg;
+        break;
       case 't':
         timeout = atoi(optarg);
         if (timeout < 0) ERRDIEX("timeout must be 0 or greater");
         break;
-      case 'q':
-        rule.queue = optarg;
-        break;
-      case 'd':
-        debug++;
-        break;
-      case 'i':
-        pidfile = optarg;
-        break;
       case 'u':
         user = optarg;
         break;
@@ -401,13 +449,6 @@ int main(int argc, char **argv)
   if (bind(sock, (struct sockaddr*) &sin, sizeof(struct sockaddr_in)) < 0)
     ERRDIE("bind()");
 
-  if (debug <= 1) {
-    /* drop privileges */
-    if (setgid(gid) != 0) ERRDIE("setgid(%lu)", (long unsigned)gid);
-    if (setuid(uid) != 0) ERRDIE("setuid(%lu)", (long unsigned)uid);
-    if (chdir("/") != 0) ERRDIE("chdir(/)");
-  }
-
   /* allocate ring buffer */
   rules = malloc(rule_count * sizeof(time_t));
   if (rules == NULL) ERRDIE("cannot allocate rules buffer");
@@ -418,7 +459,7 @@ int main(int argc, char **argv)
 
   EV_SET(&kev, sock, EVFILT_READ, EV_ADD, 0, 0, NULL);
   if (kevent(kq, &kev, 1, NULL, 0, NULL) == -1)
-    ERRDIE("cannot add socket to kqueue");
+    ERRDIE("cannot add divert socket to kqueue");
 
   /* set up signal handlers */
   memset(&sa, 0, sizeof(sa));
@@ -426,13 +467,49 @@ int main(int argc, char **argv)
   if (sigaction(SIGTERM, &sa, NULL) < 0)
     ERRDIE("sigaction(SIGTERM)");
 
+  /* limit kqueue and divert socket via capsicum */
+  cap_rights_init(&rights, CAP_KQUEUE_EVENT);
+  if (caph_rights_limit(kq, &rights) != 0)
+    ERRDIE("cannot caph_rights_limit(KQUEUE_EVENT) on kqueue");
+
+  cap_rights_init(&rights, CAP_EVENT, CAP_RECV, CAP_SEND, CAP_CONNECT);
+  if (caph_rights_limit(sock, &rights) != 0)
+    ERRDIE("cannot caph_rights_limit(EVENT,RECV,SEND,CONNECT) on divert "
+                                                                    "socket");
+
   if (debug <= 1) {
+    if (chdir("/") != 0) ERRDIE("chdir(/)");
+
     /* open syslog */
+    capcas = cap_init();
+    if (capcas == NULL) ERRDIE("cap_init()");
+
+    capsyslog = cap_service_open(capcas, "system.syslog");
+    if (capsyslog == NULL) ERRDIE("cap_service_open(system.syslog)");
+
+    cap_close(capcas);
     close(2);
-    openlog("ftp6proxy", LOG_PID, LOG_DAEMON);
+
+    cap_openlog(capsyslog, "ftp6proxy", LOG_PID | LOG_NDELAY, facility);
     log_to_syslog = 1;
+
+    /* drop privileges */
+    if (chroot(chroot_dir) != 0) ERRDIE("chroot(%s)", chroot_dir);
+    if (setgid(gid) != 0) ERRDIE("setgid(%lu)", (long unsigned)gid);
+    if (setuid(uid) != 0) ERRDIE("setuid(%lu)", (long unsigned)uid);
+  } else {
+    /* limit stderr via capsicum */
+    cap_rights_init(&rights, CAP_WRITE);
+    if (caph_rights_limit(2, &rights) != 0)
+      ERRDIE("cannot caph_rights_limit(WRITE) on stderr");
   }
 
+  /* limit via capsicum */
+/* XXX not yet possible due to sendto() with non-NULL destination XXX
+  if (caph_enter() != 0)
+    ERRDIE("caph_enter()");
+*/
+
   LOGINFOX("starting...");
 
   while (running) {
@@ -485,7 +562,8 @@ int main(int argc, char **argv)
         case 6:
           parse_v6(buf, recvlen, &rule);
           break;
-        default: break;
+        default:
+          break;
       }
     }
 
@@ -528,7 +606,7 @@ int main(int argc, char **argv)
 
   /* bye bye */
   LOGINFOX("exiting...");
-  if (log_to_syslog) closelog();
+  if (log_to_syslog) cap_closelog(capsyslog);
 
   return 0;
 }
