@@ -12,9 +12,12 @@ struct rule {
   uint16_t src_port_high;
   uint16_t dst_port_low;
   uint16_t dst_port_high;
-  char *queue;
+  char *pf_queue;
+  char *pf_anchor;
+  unsigned long ipfw_call;
 };
 
+const char *fw_ident ();
 void *fw_init ();
 int fw_close (void *data);
 int fw_add_rule (void *data, uint32_t rulenum, struct rule *rule);
