@@ -2,25 +2,52 @@
 #include <stdlib.h>
 #include <string.h>
 #include <strings.h>
+#include <errno.h>
 #include <err.h>
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <net/if.h>
 #include <netinet/ip_fw.h>
+#include <capsicum_helpers.h>
 #include "fw.h"
 
+const char *fw_ident ()
+{
+  return "ipfw";
+}
+
 void *fw_init ()
 {
+  int save_errno;
   int *ipfw_sock;
+  cap_rights_t rights;
 
   ipfw_sock = (int*) malloc(sizeof(*ipfw_sock));
-  if (ipfw_sock == NULL) return NULL;
+  if (ipfw_sock == NULL)
+    goto FW_INIT_ERR0;
 
   *ipfw_sock = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
-  if (*ipfw_sock != -1) return ipfw_sock;
+  if (*ipfw_sock < 0)
+    goto FW_INIT_ERR1;
 
+  cap_rights_init(&rights, CAP_GETSOCKOPT, CAP_SETSOCKOPT);
+  if (caph_rights_limit(*ipfw_sock, &rights) != 0)
+    goto FW_INIT_ERR2;
+
+  return ipfw_sock;
+
+FW_INIT_ERR2:
+  save_errno = errno;
+  close(*ipfw_sock);
+  errno = save_errno;
+
+FW_INIT_ERR1:
+  save_errno = errno;
   free(ipfw_sock);
+  errno = save_errno;
+
+FW_INIT_ERR0:
   return NULL;
 }
 
@@ -101,15 +128,24 @@ int fw_add_rule (void *data, uint32_t rulenum, struct rule *rule)
   port_cmd->ports[1] = rule->dst_port_high;
   cmd += 2;
 
-  cmd->opcode = O_KEEP_STATE;
-  cmd->len = 1;
-  cmd++;
+  if (rule->ipfw_call) {
+    fwrule->act_ofs = (uint32_t*) cmd - (uint32_t*) fwrule->cmd;
 
-  fwrule->act_ofs = (uint32_t*) cmd - (uint32_t*) fwrule->cmd;
+    cmd->opcode = O_CALLRETURN;
+    cmd->len = 1;
+    cmd->arg1 = rule->ipfw_call;
+    cmd++;
+  } else {
+    cmd->opcode = O_KEEP_STATE;
+    cmd->len = 1;
+    cmd++;
 
-  cmd->opcode = O_ACCEPT;
-  cmd->len = 1;
-  cmd++;
+    fwrule->act_ofs = (uint32_t*) cmd - (uint32_t*) fwrule->cmd;
+
+    cmd->opcode = O_ACCEPT;
+    cmd->len = 1;
+    cmd++;
+  }
 
   fwrule->cmd_len = (uint32_t*) cmd - (uint32_t*) fwrule->cmd;
   rulesize = RULESIZE(fwrule);
