@@ -17,7 +17,8 @@ int main()
     65535,
     22,
     22,
-    NULL
+    "test_qos_class",
+    65000
   };
 
   fw_data = fw_init();
